class LDMMaintenanceKeyError(Exception):
    """
    Exception raised when LDM Maintenance Key Error occurs
    """


class LDMServiceKeyError(Exception):
    """
    Exception raised when LDM Service Key Error occurs
    """


class LDMDatabaseKeyError(Exception):
    """
    Excepetion raised when LDM Database Key Error occurs
    """
